﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;
#if UNITY_EDITOR
using UnityEditor;
#endif

//public struct Tile {
//	int x, y;
//	public Tile( int x, int y) {
//		this.x = x;
//		this.y = y;
//    }
//}

public class Dirt : LittleMissBehaviour
{
	const int GRID_COLS = 80, GRID_ROWS = 45;
	public const int TILE_EATEN_ID = 10;
    public Transform botLeft, topRight;
	public Color[] tileDebugColors;

    // Static singleton property
    public static Dirt Instance { get; private set; }

    Rect dirtRect;
	int[][] grid; //normal coordinate direction - 0,0 is in bottom left corner
	int[] score;
	float tileWidth, tileHeight;
	List<Obstacle> obstacles = new List<Obstacle>(5);

	void Awake() {
		// First we check if there are any other instances conflicting
		if (Instance != null && Instance != this) {
			//destroy the new instance
			Destroy(this);
			return;
		}
		// Here we save our singleton instance
		Instance = this;

		// Furthermore we make sure that we don't destroy between scenes (this is optional)
		//		DontDestroyOnLoad(gameObject);

		Init();
	}

	public void Init() {
		dirtRect = new Rect();
		dirtRect.xMin = botLeft.position.x;
		dirtRect.xMax = topRight.position.x;
		dirtRect.yMin = botLeft.position.y;
		dirtRect.yMax = topRight.position.y;

		grid = new int[GRID_COLS][];
		for( int x = 0; x < GRID_COLS; x++ ) {
			grid[x] = new int[GRID_ROWS];
			for ( int y = 0; y < GRID_ROWS; y++ ) {
				grid[x][y] = -1;
            }
        }

		tileWidth = dirtRect.width / (float)GRID_COLS;
		tileHeight = dirtRect.height / (float)GRID_ROWS;

		score = new int[GameState.MAX_PLAYERS];
		for ( int i = 0; i < score.Length; i++ ) {
			GameGUI.Instance.SetScore((PlayerNum)i, 0);
        }
	}

	private void Start() {
		obstacles.AddRange(LevelState.Instance.GetComponentsInChildren<Obstacle>());
	}

	public bool IsInside( Vector2 pos ) {
        return dirtRect.Contains(pos);
    }

	public Obstacle IsInObstacle( Vector2 pos ) {
		foreach( Obstacle o in obstacles ) {
			if ( o.IsInside( pos ) ) {
				return o;
            }
        }
		return null;
    }

	Vector2 GetTileCoord(Vector2 pos) {
		//bring it into grid coord system
		pos = pos - (Vector2)botLeft.position;
		int tileX = Mathf.FloorToInt(pos.x / tileWidth);
		int tileY = Mathf.FloorToInt(pos.y / tileHeight);
		return new Vector2( tileX, tileY );
    }

	int GetTileOwner( Vector2 tileCoord ) {
		if ( (int)tileCoord.x >= GRID_COLS || (int)tileCoord.y >= GRID_ROWS  || (int)tileCoord.x < 0 || (int)tileCoord.y < 0) {
			return -2;
        }
		return grid[(int)tileCoord.x][(int)tileCoord.y];
    }

	void SetTile( Vector2 tile, PlayerNum playerNum ) {
		grid[(int)tile.x][(int)tile.y] = (int)playerNum;
	}

	void SetTileEaten(Vector2 tile) {
		grid[(int)tile.x][(int)tile.y] = TILE_EATEN_ID;
	}

	//returns the previous player num (if exists)
	//10 = ate a veggie
	//0-4 = prev owner
	//-1 = nobody owned
	//-2 = bogus
	public int UpdateWormSegmentPos( PlayerNum playerNum, Vector2 pos, bool head, bool tail, bool isEating ) {
		Vector2 tileCoord = GetTileCoord(pos);

		int tileOwner = GetTileOwner(tileCoord);
		if (tileOwner == -2) { //means bogus coords
			return -2;
		}
		//EAT
		if (isEating && head && tileOwner == -1 ) {
			SetTileEaten(tileCoord);
			return TILE_EATEN_ID;
		}
		else if (tileOwner != (int)playerNum && tileOwner != TILE_EATEN_ID) {
			if (tail) {
				SetTile(tileCoord, playerNum);
				//handle score - decrement if another player owns it, and increment the new player
				if (tileOwner >= 0) {
					score[tileOwner]--;
					GameGUI.Instance.SetScore((PlayerNum)tileOwner, score[tileOwner]);
				}
				score[(int)playerNum]++;
				GameGUI.Instance.SetScore(playerNum, score[(int)playerNum]);
			}
			return tileOwner; //could be -1 (nobody owns) or a player num
		}
		return (int)playerNum; //means prev owner is self
    }

#if UNITY_EDITOR
    private void OnDrawGizmos() {
		if ( Application.isPlaying ) {
			for (int x = 0; x < GRID_COLS; x++) {
				for (int y = 0; y < GRID_ROWS; y++) {
					if ( grid[x][y] > tileDebugColors.Length - 1 ) {
						continue;
					}
					Gizmos.color = grid[x][y] < 0 ? Color.gray : tileDebugColors[grid[x][y]];
					Vector2 center = (Vector2)botLeft.position + new Vector2( (x) * tileWidth, ( y ) * tileHeight );
					Vector2 size = new Vector2(tileWidth, tileHeight) * .98f; //tiny bit smaller so edges don't overlap
					DrawRect(new Rect(center, size));
					//or could use Gizmos.DrawCube() maybe?
				}
			}
		}
		
		
	}

    void DrawRect(Rect rect) {
		Gizmos.DrawWireCube(new Vector3(rect.center.x, rect.center.y, 0.01f), new Vector3(rect.size.x, rect.size.y, 0.01f));
	}
#endif
}
