﻿using UnityEngine;
using UnityEngine.UI;
using System;
using System.Collections;
using System.Collections.Generic;

public class Line {
	public Vector3 start, end;
	public Color color;
	public float time;
	public Line( Vector3 _start, Vector3 _end, Color _color, float _time ) {
		start = _start;
		end = _end;
		color = _color;
		time = _time;
	}
	public bool Expired() {
		time -= Time.deltaTime;
		return time <= 0f;
	}
}

public static class Util {
	#region DEBUG
	public static void DrawLine( Vector3 _start, Vector3 _end, Color _color, float _time = 0f ) {
		Debug.DrawLine( _start, _end, _color, _time );
	}

	public static void DrawRect( Vector3 bottomLeft, Vector3 bottomRight, Vector3 topRight, Vector3 topLeft, Color _color, float _time = 0f ) {
		Util.DrawLine( bottomLeft, bottomRight, _color, _time );
		Util.DrawLine( bottomRight, topRight, _color, _time );
		Util.DrawLine( topRight, topLeft, _color, _time );
		Util.DrawLine( topLeft, bottomLeft, _color, _time );
	}

	public static void DrawRect( Vector3 bottomLeft, Vector3 topRight, Color _color, float _time = 0f ) {
		float width = topRight.x - bottomLeft.x;
		Util.DrawLine( topRight, bottomLeft + Vector3.right * width, _color, _time );//right
		Util.DrawLine( bottomLeft + Vector3.right * width, bottomLeft, _color, _time );//bottom
		Util.DrawLine( bottomLeft, topRight - Vector3.right * width, _color, _time );//left
		Util.DrawLine( topRight - Vector3.right * width, topRight, _color, _time );//top
	}

	public static void DrawRect( Bounds bounds, Color _color, float _time = 0f ) {
		DrawRect( bounds.min, bounds.max, _color, _time );
	}

	public static void DrawDiamond( Vector3 center, float size, Color _color, float _time = 0f ) {
		Util.DrawCircle( center, size, _color, _time, 2f * Mathf.PI / 4f );
	}

	//TERRIBLE HACK TO DRAW DEBUG CIRCLES :(
	public static void DrawCircle( Vector3 center, float radius, Color _color, float _time = 0f, float _intervalOverride = -1f ) {
		float interval = _intervalOverride > 0f ? _intervalOverride : .2f;
		Vector3 pos = center, prevPos = center + Vector3.right * radius;
		for ( float theta = interval; theta < 2f * Mathf.PI; theta += interval ) {
			pos.x = center.x + radius * Mathf.Cos( theta );
			pos.y = center.y + radius * Mathf.Sin( theta );
			Util.DrawLine( prevPos, pos, _color, _time );
			prevPos = pos;
		}
		//connect back to start
		Util.DrawLine( pos, center + Vector3.right * radius, _color, _time );
	}


	public static void Print( float msg ) {
		Print( msg.ToString() );
	}

	public static void Print( bool msg ) {
		Print( msg.ToString() );
	}

	public static void Print( int msg ) {
		Print( msg.ToString() );
	}

	public static void Print( Vector3 msg ) {
		Print( msg.ToString() );
	}

	public static void Print( string msg ) {
		#if !UNITY_EDITOR
			return;
		#endif
		Debug.Log( msg );
	}

	//taken from http://forum.unity3d.com/threads/assert-in-unity3d.44513/
	//be careful - string concats are expensive in this
	[System.Diagnostics.Conditional( "UNITY_EDITOR" )]
	[System.Diagnostics.Conditional( "DEBUG" )]
	public static void Assert( bool comparison, string message = "" ) {
		if ( !comparison ) {
			//throw new UnityException( message );
			message = "ASSERT! " + message;
			Debug.LogWarning( message );
			Debug.Break();
		}
	}

	//be careful - string concats are expensive in this
	[System.Diagnostics.Conditional( "UNITY_EDITOR" )]
	[System.Diagnostics.Conditional( "DEBUG" )]
	public static void Warning( bool comparison, string message = "" ) {
		if ( !comparison ) {
			message = "WARNING! " + message;
			Debug.LogWarning( message );
		}
	}
	#endregion

	#region COLLISION/INTERSECTION
	//from http://answers.unity3d.com/questions/163864/test-if-point-is-in-collider.html
	public static bool IsInside ( Collider test, Vector3 point)
	{
		Vector3 center;
		Vector3 direction;
		Ray ray;
		RaycastHit hitInfo;
		bool hit;
		
		// Use collider bounds to get the center of the collider. May be inaccurate
		// for some colliders (i.e. MeshCollider with a 'plane' mesh)
		center = test.bounds.center;
		
		// Cast a ray from point to center
		direction = center - point;
		ray = new Ray(point, direction);
		hit = test.Raycast(ray, out hitInfo, direction.magnitude);
		
		// If we hit the collider, point is outside. So we return !hit
		return !hit;
	}

	//from http://forum.unity3d.com/threads/line-intersection.17384/
	public static bool LineIntersection( Vector2 p1, Vector2 p2, Vector2 p3, Vector2 p4, ref Vector2 intersection ) {
		float Ax, Bx, Cx, Ay, By, Cy, d, e, f, num;
		float x1lo, x1hi, y1lo, y1hi;
		Ax = p2.x - p1.x;
		Bx = p3.x - p4.x;

		// X bound box test/	
		if ( Ax < 0 ) {
			x1lo = p2.x; x1hi = p1.x;
		}
		else {
			x1hi = p2.x; x1lo = p1.x;
		}

		if ( Bx > 0 ) {
			if ( x1hi < p4.x || p3.x < x1lo ) return false;
		}
		else {
			if ( x1hi < p3.x || p4.x < x1lo ) return false;
		}

		Ay = p2.y - p1.y;
		By = p3.y - p4.y;

		// Y bound box test//
		if ( Ay < 0 ) {
			y1lo = p2.y; y1hi = p1.y;
		}
		else {
			y1hi = p2.y; y1lo = p1.y;
		}

		if ( By > 0 ) {
			if ( y1hi < p4.y || p3.y < y1lo ) return false;
		}
		else {
			if ( y1hi < p3.y || p4.y < y1lo ) return false;
		}

		Cx = p1.x - p3.x;
		Cy = p1.y - p3.y;
		d = By * Cx - Bx * Cy;  // alpha numerator//
		f = Ay * Bx - Ax * By;  // both denominator//

		// alpha tests//
		if ( f > 0 ) {
			if ( d < 0 || d > f ) return false;
		}
		else {
			if ( d > 0 || d < f ) return false;
		}

		e = Ax * Cy - Ay * Cx;  // beta numerator//

		// beta tests //
		if ( f > 0 ) {
			if ( e < 0 || e > f ) return false;
		}
		else {
			if ( e > 0 || e < f ) return false;
		}

		// check if they are parallel
		if ( f == 0 ) return false;
		// compute intersection coordinates //
		num = d * Ax; // numerator //
		intersection.x = p1.x + num / f;

		num = d * Ay;
		intersection.y = p1.y + num / f;
		return true;
	}

	//derived from http://stackoverflow.com/questions/1073336/circle-line-collision-detection
	public static bool LineIntersectsCircle( Vector2 p1, Vector2 p2, Vector2 circlePoint, float circleRadius ) {
		float circleRadiusSqr = circleRadius * circleRadius;
		if ( ( circlePoint - p1 ).sqrMagnitude <= circleRadiusSqr || ( circlePoint - p2 ).sqrMagnitude <= circleRadiusSqr ) {
			return true;
		}
		//project p1->circlePoint onto p1->p2
		Vector2 projectedPoint = p1 + (Vector2)Vector3.Project( circlePoint - p1, p2 - p1 );
		//DEBUG
		//Util.DrawLine( p1, p2, Color.blue, 1f );
		//Util.DrawLine( circlePoint, projectedPoint, Color.red, 1f );
		//make sure point lies on line segment (b/w p1 and p2)
		if ( projectedPoint.x > Mathf.Max( p1.x, p2.x ) || projectedPoint.x < Mathf.Min( p1.x, p2.x )
			|| projectedPoint.y > Mathf.Max( p1.y, p2.y ) || projectedPoint.y < Mathf.Min( p1.y, p2.y ) ) {
				//Util.DrawLine( circlePoint, projectedPoint, Color.cyan, 1f );
			return false;
		}
		return ( circlePoint - projectedPoint ).sqrMagnitude <= circleRadiusSqr;
	}

	#endregion

	#region MATH
	public static float RoundToNum( float amount, float roundToNum ) {
		return Mathf.Round( amount / roundToNum ) * roundToNum;
	}

	public static bool Cointoss() {
		return UnityEngine.Random.value < 0.5f;
	}

	public static int RandomSign() {
		if ( Cointoss() ) {
			return -1;
		}
		return 1;
	}

	public static bool Approximately( float a, float b, float epsilon ) {
		return Mathf.Abs( a - b ) <= epsilon;
	}

	//assumes 2D XY view
	public static Vector3 AngleToUnitVector( float angle, bool alignAngleToTop = true ) {
		if ( alignAngleToTop ) {
			//add 90 to align coord systems to be 0 deg at top (rather than right)
			angle = ( angle + 90f ) * Mathf.Deg2Rad;
		}
		return new Vector3( Mathf.Cos( angle ), Mathf.Sin( angle ), 0f );
	}

	public static float VectorToAngle( Vector2 vec ) {
		return -Vector2.Angle( Vector2.up, vec ) * Mathf.Sign( Vector2.Dot( vec, Vector2.right ) );
	}

	public static Quaternion VectorToQuaternion( Vector3 vector ) {
		return Quaternion.FromToRotation( Vector3.forward, vector );
	}

	public static Vector3 RotatePointAroundPivot( Vector3 point, Vector3 pivot, Vector3 angles ) {
		Vector3 dir = point - pivot; // get point direction relative to pivot
		dir = Quaternion.Euler( angles ) * dir; // rotate it
		point = dir + pivot; // calculate rotated point
		return point; // return it
	}


	public static float DistToRay( Vector3 lineStart, Vector3 lineEnd, Vector3 point ) {
		return DistToRay( new Ray( lineStart, lineEnd - lineStart ), point );
	}

	public static float DistToRay( Vector4 line, Vector3 point ) {
		Vector2 vector1 = new Vector2( line.x, line.y );
		Vector2 vector2 = new Vector2( line.z, line.w );
		return DistToRay( new Ray( vector1, vector2 - vector1 ), point );
	}

	private static float DistToRay( Ray ray, Vector3 point ) {
		return Vector3.Cross( ray.direction, point - ray.origin ).magnitude;
	}

	//TODO - ugly
	public static float DistToLine( Vector2 lineStart, Vector2 lineEnd, Vector2 point ) {
		float distToRay = DistToRay (lineStart, lineEnd, point);
		float distToLineStart = Vector2.Distance (point, lineStart);
		float distToLineEnd = Vector2.Distance (point, lineEnd);
		if ( IsWithinLineSegment( point, new Vector4( lineStart.x, lineStart.y, lineEnd.x, lineEnd.y ) ) ) {
			return distToRay;
		}
		return Mathf.Min (distToLineStart, distToLineEnd);
	}


	//EXTENSION
	public static Vector2 PerpindicularUnitVector( this Vector2 v, bool left = true ) {
		Vector2 perp = v.normalized;
		if ( left ) {
			float tempX = perp.x;
			perp.x = -perp.y;
			perp.y = tempX;
		}
		else {
			float tempY = perp.y;
			perp.y = -perp.x;
			perp.x = tempY;
		}
		return perp;
	}

	//EXTENSION
	public static bool Within( this float f, float min, float max ) {
		return f >= min && f <= max;
	}

	//EXTENSION
	public static bool WithinExclusive( this float f, float min, float max ) {
		return f > min && f < max;
	}

	//Extension
	//efficiently remove an item from an array, by swapping the last item into the specified index, and then downsizing by 1.
	public static void RemoveSwapLastAt<T>( this List<T> list, int index ) {
		list[index] = list[list.Count-1];
		list.RemoveAt(list.Count-1);
	}

	public static void RemoveSwapLast<T>( this List<T> list, T item ) {
		int childIdx = list.IndexOf(item);
		if(childIdx>=0){
			list[childIdx] = list[list.Count-1];
			list.RemoveAt(list.Count-1);
		}
	}

	//from https://forum.unity.com/threads/limiting-rotation-with-mathf-clamp.171294/#post-2244265
	//don't quite understand it but so far it works!
	public static float ClampAngle(float angle, float min, float max) {
		if (min < 0 && max > 0 && (angle > max || angle < min)) {
			angle -= 360;
			if (angle > max || angle < min) {
				if (Mathf.Abs(Mathf.DeltaAngle(angle, min)) < Mathf.Abs(Mathf.DeltaAngle(angle, max))) return min;
				else return max;
			}
		}
		else if (min > 0 && (angle > max || angle < min)) {
			angle += 360;
			if (angle > max || angle < min) {
				if (Mathf.Abs(Mathf.DeltaAngle(angle, min)) < Mathf.Abs(Mathf.DeltaAngle(angle, max))) return min;
				else return max;
			}
		}
		if (angle < min) return min;
		else if (angle > max) return max;
		else return angle;
	}

	//get distance to ray from pt and the vector to the ray (always perpindicular)
	public static Vector2 ClampToRay( Vector2 pos, Vector4 ray ) {
		Vector2 rayStart = new Vector2( ray.x, ray.y );
		Vector2 rayDirectionPoint = new Vector2( ray.z, ray.w );
		Vector2 lineVector = rayDirectionPoint - rayStart;
		Vector2 perpVector = lineVector.PerpindicularUnitVector();

		float dist = DistToRay( rayStart, rayDirectionPoint, pos );
		if ( Vector2.Dot( perpVector, pos - rayStart ) > 0f ) { //which side of ray are we on?
			dist *= -1f;
		}
		//Debug.Log( pos + " " + perpVector + " " + dist );
		return pos + perpVector * dist;
	}

	public static Vector2 ClampToLine( Vector2 pos, Vector4 line ) {
		Vector2 vector1 = new Vector2( line.x, line.y );
		Vector2 vector2 = new Vector2( line.z, line.w );
		Vector2 lineVector = vector2 - vector1;
		Vector2 reverseLineVector = vector1 - vector2;
		Vector2 perpVector = lineVector.PerpindicularUnitVector();

		//too far to local left
		if ( Vector2.Dot( lineVector, pos - vector1 ) < 0f ) {
			return vector1;
		}
		//too far to local right
		if ( Vector2.Dot( reverseLineVector, pos - vector2 ) < 0f ) {
			return vector2;
		}
		
		float dist = DistToRay( vector1, vector2, pos );
		if ( Vector2.Dot( perpVector, pos - vector1 ) > 0f ) {
			dist *= -1f;
		}
		//Debug.Log( pos + " " + perpVector + " " + dist );
		return pos + perpVector * dist;
	}

	public static bool IsWithinLineSegment( Vector2 pos, Vector4 line ) {
		Vector2 vector1 = new Vector2( line.x, line.y );
		Vector2 vector2 = new Vector2( line.z, line.w );
		Vector2 lineVector = vector2 - vector1;
		Vector2 reverseLineVector = vector1 - vector2;

		//too far to local left
		if ( Vector2.Dot( lineVector, pos - vector1 ) < 0f ) {
			return false;
		}
		//too far to local right
		if ( Vector2.Dot( reverseLineVector, pos - vector2 ) < 0f ) {
			return false;
		}
		return true;
	}


	public static int Sum( int[] ints ) {
		int result = 0;
		for ( int i = 0; i < ints.Length; i++ ) {
			result += ints[i];
		}
		return result;
	}

	public static float Sum( float[] floats ) {
		float result = 0f;
		for ( int i = 0; i < floats.Length; i++ ) {
			result += floats[i];
		}
		return result;
	}

	public static float Average( float[] list ) {
		int len = list.Length;
		float sum = 0f;
		for ( int i = 0; i < len; i++ ) {
			sum += list[i];
		}
		return sum / len;
	}

	//histogram - for weighted random distributions - returns index it randomly selects
	//total of distro[] can be > 1
	public static int WeightedRandom( float[] distro ) {
		float total = Sum( distro );
		float randVal = UnityEngine.Random.value * total;
		float currVal = 0f;
		for ( int i = 0; i < distro.Length; i++ ) {
			currVal += distro[i];
			if ( randVal < currVal ) {
				return i;
			}
		}
		return distro.Length - 1;
	}
	#endregion

	#region MOBILE
	public static bool isPhone = false;
	//TODO - maybe eventually this returns a float for how much of a phone it is :)
	public static bool IsPhone() {
		if ( isPhone ) {
			return true;
		}
#if UNITY_IOS
		isPhone = UnityEngine.iOS.Device.generation.ToString().IndexOf( "iPhone" ) > -1;		
#else
		float widthInInches = Screen.width / Screen.dpi;
		float heightInInches = Screen.height / Screen.dpi;
		float diagInInchesSq = widthInInches * widthInInches + heightInInches * heightInInches;
		isPhone = diagInInchesSq <= 5.0f * 5.0f;

		#endif
		return isPhone;
	}

	public static bool IsiPad() {
	#if UNITY_IOS
		return UnityEngine.iOS.Device.generation.ToString().IndexOf( "iPad" ) > -1;
	#else
		return false;
	#endif
	}
	#endregion

	#region MISC
	public static Color SetAlpha( Color color, float alpha ) {
		Color c = color;
		c.a = alpha;
		return c;
	}

	public static string RemoveCloneSuffix( this string name ) {
		return name.Replace( "(Clone)", "" );
	}

	public static IEnumerator WaitForRealTime( float delay ) {
		while ( true ) {
			float pauseEndTime = Time.realtimeSinceStartup + delay;
			while ( Time.realtimeSinceStartup < pauseEndTime ) {
				yield return null;
			}
			break;
		}
	}
	#endregion




	#region EXTENSIONS
	public static void SetPosX( this RectTransform t, float newX ) {
		t.anchoredPosition = new Vector2( newX, t.anchoredPosition.y );
	}

	public static void SetPosY( this RectTransform t, float newY ) {
		t.anchoredPosition = new Vector2( t.anchoredPosition.x, newY );
	}


	public static void SetRotZ( this Transform t, float newZ ) {
		t.eulerAngles = new Vector3( t.eulerAngles.x, t.eulerAngles.y, newZ );
	}

	public static void SetX(this Transform t, float newX)
	{
		t.position = new Vector3(newX, t.position.y, t.position.z);
	}
	
	public static void SetY(this Transform t, float newY)
	{
		t.position = new Vector3(t.position.x, newY, t.position.z);
	}
	
	public static void SetZ(this Transform t, float newZ)
	{
		t.position = new Vector3(t.position.x, t.position.y, newZ);
	}

	public static void SetLocalPosX(this Transform t, float newX)
	{
		t.localPosition = new Vector3(newX, t.localPosition.y, t.localPosition.z);
	}
	
	public static void SetLocalPosY(this Transform t, float newY)
	{
		t.localPosition = new Vector3(t.localPosition.x, newY, t.localPosition.z);
	}
	
	public static void SetLocalPosZ(this Transform t, float newZ)
	{
		t.localPosition = new Vector3(t.localPosition.x, t.localPosition.y, newZ);
	}

	public static void SetScaleX(this Transform t, float newX)
	{
		t.localScale = new Vector3(newX, t.localScale.y, t.localScale.z);
	}
	
	public static void SetScaleY(this Transform t, float newY)
	{
		t.localScale = new Vector3(t.localScale.x, newY, t.localScale.z);
	}
	
	public static void SetScaleZ(this Transform t, float newZ)
	{
		t.localScale = new Vector3(t.localScale.x, t.localScale.y, newZ);
	}

	public static void Shuffle<T>(this List<T> list)  
	{  
		System.Random rng = new System.Random();  
		int n = list.Count;  
		while (n > 1) {  
			n--;  
			int k = rng.Next(n + 1);  
			T value = list[k];  
			list[k] = list[n];  
			list[n] = value;  
		}  
	}

	public static void Shuffle<T>(this T[] arr)  
	{  
		System.Random rng = new System.Random();  
		int n = arr.Length;  
		while (n > 1) {  
			n--;  
			int k = rng.Next(n + 1);  
			T value = arr[k];  
			arr[k] = arr[n];  
			arr[n] = value;  
		}  
	}

	public static void Each<T>( this IEnumerable<T> ie, Action<T, int> action )
	{
		int i = 0;
		foreach ( T e in ie ) action( e, i++ );
	}
	//EXAMPLE USAGE:
//	List<string> strings = new List<string>();
//	strings.Each( ( str, n ) =>
//	{
//		// hooray
	//	} );


	//******Orthographic Camera Only******//
	public static Vector2 BoundsMin( this Camera camera ) {
		return (Vector2)camera.transform.position - camera.Extents();
	}

	public static Vector2 BoundsMax( this Camera camera ) {
		return (Vector2)camera.transform.position + camera.Extents();
	}

	public static Vector2 Extents( this Camera camera ) {
		if ( camera.orthographic )
			return new Vector2( camera.orthographicSize * Screen.width / Screen.height, camera.orthographicSize );
		else {
			Debug.LogError( "Camera is not orthographic!", camera );
			return new Vector2();
		}
	}

	public static Bounds GetOrthographicBounds( this Camera camera ) {
		return new Bounds( camera.transform.position, camera.BoundsMax() - camera.BoundsMin() );
	}
	//*****End of Orthographic Only*****//

	//vertex color stuff
	public static Color32[] colors;
	public static void SetAllVertexColors( this MeshFilter meshFilter, Color32 color, ref Color32[] meshColors  ) {
		Mesh mesh = meshFilter.mesh;
		int i = 0;
		while ( i < mesh.vertexCount ) {
			meshColors[i] = color;
			i++;
		}
		mesh.colors32 = meshColors;
	}

	public static void SetColor( this Button button, Color color ) {
		ColorBlock cb = button.colors;
		cb.normalColor = color;
		button.colors = cb;
	}

	public static void SetHighlightColor( this Button button, Color color ) {
		ColorBlock cb = button.colors;
		cb.highlightedColor = cb.pressedColor = color;
		button.colors = cb;
	}

	public static void ToggleHighlight( this Button button ) {
		ColorBlock cb = button.colors;
		//hacky - swap colors to save the old normal color
		Color temp = cb.normalColor;
		cb.normalColor = cb.disabledColor;
		cb.disabledColor = temp;
		button.colors = cb;
	}
	#endregion
}

public class LittleMissBehaviour : MonoBehaviour {
	private GameObject _gameObject;
	public new GameObject gameObject {
		get {
			if (!_gameObject) {
				_gameObject = base.gameObject;
			}
			return _gameObject;
		}
	}
	private Transform _transform;
	public new Transform transform {
		get {
			if (!_transform) {
				_transform = base.transform;
			}
			return _transform;
		}
	}
	private Renderer _renderer;
	public new Renderer renderer {
		get {
			if (!_renderer) {
				_renderer = base.GetComponent<Renderer>();
			}
			return _renderer;
		}
	}
	private SpriteRenderer _spriteRenderer;
	public SpriteRenderer spriteRenderer {
		get {
			if (!_spriteRenderer) {
				_spriteRenderer = base.GetComponent<Renderer>() as SpriteRenderer;
			}
			return _spriteRenderer;
		}
	}
	private MeshFilter _meshFilter;
	public MeshFilter meshFilter {
		get {
			if (!_meshFilter) {
				_meshFilter = GetComponent<MeshFilter>();
			}
			return _meshFilter;
		}
	}
	private Collider _collider;
	public new Collider collider {
		get {
			if (!_collider) {
				_collider = base.GetComponent<Collider>();
			}
			return _collider;
		}
	}
	private Collider2D _collider2D;
	public new Collider2D collider2D {
		get {
			if (!_collider2D) {
				_collider2D = base.GetComponent<Collider2D>();
			}
			return _collider2D;
		}
	}
	private ParticleSystem _particleSystem;
	public new ParticleSystem particleSystem {
		get {
			if ( !_particleSystem ) {
				_particleSystem = base.GetComponent<ParticleSystem>();
			}
			return _particleSystem;
		}
	}
	private AudioSource _audio;
	public new AudioSource audio {
		get {
			if ( !_audio ) {
				_audio = base.GetComponent<AudioSource>();
			}
			return _audio;
		}
	}
	private Camera _camera;
	public new Camera camera {
		get {
			if ( !_camera ) {
				_camera = base.GetComponent<Camera>();
			}
			return _camera;
		}
	}

	private Text _text;
	public Text text {
		get {
			if ( !_text ) {
				_text = base.GetComponent<Text>();
			}
			return _text;
		}
	}
}